from typing import Optional

from pydantic import BaseModel


class UsageInfo(BaseModel):
    """
    Common because will implement future text_completions (legacy) endpoint support in future
    """

    prompt_tokens: int
    # total_tokens: int
    completion_tokens: Optional[int]
    total_charged: float



from typing import List

from pydantic import BaseModel

from shuttleai.schemas.common import UsageInfo


class EmbeddingObject(BaseModel):
    object: str
    embedding: List[float]
    index: int


class EmbeddingResponse(BaseModel):
    object: str
    data: List[EmbeddingObject]
    model: str
    usage: UsageInfo



import httpx


def _patch_httpx() -> None:
    """
    Monkey-patch httpx to accept direct bytes (orjson.dumps) directly.
    Originally from
    - https://github.com/encode/httpx/issues/717
    But modified to be faster for ShuttleAI's specific use case.
    Do not use this patch outside of ShuttleAI if you don't know what you're doing.
    """
    from httpx._content import Any, ByteStream

    def encode_json(json: Any) -> tuple[dict[str, str], ByteStream]:
        return {
            "Content-Length": str(len(json)),
            "Content-Type": "application/json",
        }, ByteStream(json)

    # This makes the above function look and act like the original.
    encode_json.__globals__.update(httpx._content.__dict__)
    encode_json.__module__ = httpx._content.__name__
    httpx._content.encode_json = encode_json



from pydantic import BaseModel


class AudioSpeech(BaseModel):
    url: str
    """The URL of the audio file."""


class AudioSpeechResponse(BaseModel):
    created: int
    """The Unix timestamp when the audio generation was created."""

    data: AudioSpeech
    """The generated audio file."""

    model: str
    """The model used to generate the audio file."""

    chars: int
    """The number of characters in the audio file."""

    expiresIn: int = 3600
    """The number of seconds before the audio file expires."""



from typing import List

from pydantic import BaseModel


class Image(BaseModel):
    url: str
    """The URL of the image."""

    def to_file(self, path: str) -> None:
        """Save the image to a file.

        Args:
            path (str): The path to save the image to.
        """
        import httpx

        response = httpx.get(self.url)
        with open(path, "wb") as file:
            file.write(response.content)

    def to_bytes(self) -> bytes:
        """Get the image as bytes.

        Returns:
            bytes: The image as bytes.
        """
        import httpx

        response = httpx.get(self.url)
        return response.content

    def show(self) -> None:
        """Show the image using pillow."""
        from io import BytesIO

        from PIL import Image as PILImage

        image = PILImage.open(BytesIO(self.to_bytes()))
        image.show()

    def __str__(self) -> str:
        return self.url


class ImagesGenerationResponse(BaseModel):
    created: int
    """The Unix timestamp when the image generation was created."""

    data: List[Image]
    """The generated image(s)."""

    @property
    def first_image(self) -> Image:
        return self.data[0]



from typing import List

from pydantic import BaseModel


class Insult(BaseModel):
    insult: str
    """The insult."""


class InsultResponse(BaseModel):
    data: List[Insult]
    """The list of insults."""



from typing import Literal, Optional, Union

from shuttleai.resources.common import AsyncResource, SyncResource
from shuttleai.schemas.etc.insults import InsultResponse


class AsyncInsults(AsyncResource):
    async def generate(
        self,
        model: Optional[Union[str, Literal["insult-1"]]] = "insult-1",
    ) -> InsultResponse:
        return await self.handle_request(  # type: ignore
            method="get",
            endpoint=f"v1/insults?key={self._client.api_key}&model={model}",
            request_data=None,
            response_cls=InsultResponse,
        )


class Insults(SyncResource):
    def generate(
        self,
        model: Optional[Union[str, Literal["insult-1"]]] = "insult-1",
    ) -> InsultResponse:
        return self.handle_request(  # type: ignore
            method="get",
            endpoint=f"v1/insults?key={self._client.api_key}&model={model}",
            request_data=None,
            response_cls=InsultResponse,
        )



from typing import Literal, Optional, Union

from shuttleai.resources.common import AsyncResource, SyncResource
from shuttleai.schemas.etc.jokes import JokeResponse


class AsyncJokes(AsyncResource):
    async def generate(
        self,
        model: Optional[Union[str, Literal["joke-1"]]] = "joke-1",
    ) -> JokeResponse:
        return await self.handle_request(  # type: ignore
            method="get",
            endpoint=f"v1/jokes?key={self._client.api_key}&model={model}",
            request_data=None,
            response_cls=JokeResponse,
        )


class Jokes(SyncResource):
    def generate(
        self,
        model: Optional[Union[str, Literal["joke-1"]]] = "joke-1",
    ) -> JokeResponse:
        return self.handle_request(  # type: ignore
            method="get",
            endpoint=f"v1/jokes?key={self._client.api_key}&model={model}",
            request_data=None,
            response_cls=JokeResponse,
        )



from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel

from shuttleai.schemas.models.capabilities import Capabilities


class ProxyType(str, Enum):
    proxy = "proxy"


class ModelType(str, Enum):
    model = "model"


class BaseModelCard(BaseModel):
    id: str
    """The ID of the model object."""

    object: ModelType = ModelType.model
    """The object type, which is always "model" for a model object."""

    created: int
    """The Unix timestamp when the model was created."""

    owned_by: str
    """The organization ID of the owner of the model."""

    cost: int = 1
    """The usage cost per request to this model."""

    premium: bool = False
    """Whether the model requires premium permissions."""

    endpoint: Union[str, List[str]]
    """The endpoint URL(s) for the model."""


class VerboseModelCard(BaseModelCard):
    name: str
    """The pretty name of the model."""

    description: str
    """The description of the model."""

    maintenance: Optional[bool] = None
    """Whether the model is under maintenance. If true, the model is not available for use."""

    beta: Optional[bool] = None
    """Whether the model is in beta. If true, the model may have unstable behavior and must be requested to use."""

    capabilities: Optional[Capabilities] = None
    """The capabilities of the model."""


class ProxyCard(BaseModel):
    id: str
    """The ID of the proxy object."""

    object: ProxyType = ProxyType.proxy
    """The object type, which is always "proxy" for a proxy object."""

    proxy_to: str
    """The model ID that the proxy points to."""

    @property
    def parent(self) -> BaseModelCard | str:
        """The parent model card that the proxy points to."""
        return f"tmp-{self.proxy_to}"

    @parent.setter
    def parent(self, value: BaseModelCard) -> None:
        self._parent = value


ListModelsData = List[Union[BaseModelCard, ProxyCard]]


ListVerboseModelsData = List[Union[VerboseModelCard, ProxyCard]]


class ListModelsResponse(BaseModel):
    object: str
    """The object type, which is always "list" for a list of model objects."""

    data: ListModelsData
    """The list of model/proxy cards."""

    count: int
    """The number of model/proxy cards returned."""


class ListVerboseModelsResponse(BaseModel):
    object: str
    """The object type, which is always "list" for a list of model objects."""

    data: ListVerboseModelsData
    """The list of model/proxy cards."""

    count: int
    """The number of model/proxy cards returned."""



from typing import List, Union

from pydantic import BaseModel


class WebSearchResult(BaseModel):
    title: str
    """The title of the search result."""

    link: str
    """The link to the search result."""

    snippet: str
    """The snippet of the search result."""


class WebSearchImageResult(BaseModel):
    title: str
    """The title of the image search result."""

    link: str
    """The link to the image search result."""

    size: str
    """The size of the image search result."""

    context: str
    """The context of the image search result."""

    byteSize: str
    """The byte size of the image search result."""

    mime: str
    """The MIME type of the image search result."""


WebSearchResponses = List[WebSearchResult]
WebSearchImageResponses = List[WebSearchImageResult]


EitherWebSearchResponses = Union[WebSearchResponses, WebSearchImageResponses]

# TODO: Remake web search api format on API side to have a consistent formatting throughout
# both DDG and Google search APIs including both general web search and image search.



import inspect
import re
from typing import Callable, Literal, Type, get_type_hints


def _get_type_name(t: Type) -> str:
    """Gets the name of a type, handling some edge cases like Literal types

    Args:
        t (Type): The type to get the name of

    Returns:
        str: The name of the type
    """
    if hasattr(t, "__name__"):
        return t.__name__
    if hasattr(t, "_name"):
        assert isinstance(t._name, str)
        return t._name
    return str(t)


def serialize_function_to_json(func: Callable) -> dict:
    """Serializes a python callable function to a function schema JSON spec

    Args:
        func (Callable): The function to serialize

    Returns:
        dict: The serialized function schema JSON spec
    """
    signature = inspect.signature(func)
    type_hints = get_type_hints(func)
    params = signature.parameters

    function_info = {
        "name": func.__name__,
        "description": (func.__doc__ or "Lorem ipsum...").split(":param")[0].strip(),
        "parameters": {
            "type": "object",
            "properties": {},
        },
    }

    docstring = func.__doc__ or ""
    param_desc_pattern = r":param\s+(\w+)\s*:(.*?)(?=:param|$)"
    param_descriptions = dict(re.findall(param_desc_pattern, docstring, re.DOTALL))

    required_params = [name for name, param in params.items() if param.default == inspect.Parameter.empty]
    if required_params:
        function_info["parameters"]["required"] = required_params  # type: ignore

    for name, _ in params.items():
        param_type = _get_type_name(type_hints.get(name, type(None)))
        param_desc = param_descriptions.get(name, "Lorem ipsum...").strip()
        param_info = {"type": param_type, "description": param_desc}
        if isinstance(type_hints.get(name), type(Literal)):
            param_info["enum"] = list(type_hints[name].__args__)
            param_info["type"] = "string"
        function_info["parameters"]["properties"][name] = param_info  # type: ignore

    return function_info


def deserialize_function_from_json(json_obj: dict) -> Callable:
    """Deserializes a function schema JSON spec to a python callable function

    Args:
        json_obj (dict): The function schema JSON spec

    Returns:
        Callable: The deserialized python callable function
    """
    func_name = json_obj["name"]
    func_docstring = json_obj["description"]

    param_types = {}
    param_docstrings = {}
    for param_name, param_info in json_obj["parameters"]["properties"].items():
        param_type = param_info["type"]
        try:
            if param_type == "string" and "enum" in param_info:
                param_types[param_name] = Literal.__getitem__(tuple(param_info["enum"]))
            else:
                param_types[param_name] = eval(param_type)
        except NameError:
            param_types[param_name] = str  # Fallback type if eval fails
        param_docstrings[param_name] = param_info["description"]

    param_docstring = "\n".join([f":param {name}: {desc}" for name, desc in param_docstrings.items()])
    full_docstring = f"{func_docstring}\n{param_docstring}"

    def func(*args, **kwargs):  # type: ignore
        real_func = globals()[func_name]
        return real_func(*args, **kwargs)

    func.__name__ = func_name
    func.__doc__ = full_docstring
    func.__annotations__ = param_types

    return func


def convert_function_json_to_tool_json(json_obj: dict) -> dict:
    """Converts a serialized function JSON object to a tool JSON SPEC

    Args:
        json_obj (dict): The serialized function JSON object

    Returns:
        dict: The tool JSON SPEC object
    """
    return {"type": "function", "function": json_obj}



from typing import Any, AsyncIterable, AsyncIterator, Dict, Iterator, Type, TypeVar

from pydantic import BaseModel

from shuttleai.client.base import ClientBase
from shuttleai.exceptions import ShuttleAIException

T = TypeVar("T", bound=ClientBase)


class BaseResource:
    def __init__(self, client: ClientBase):
        self._client = client


class SyncResource(BaseResource):
    def _stream_response(
        self, response: Iterator[Dict[str, Any]], response_cls: Type[BaseModel]
    ) -> Iterator[BaseModel]:
        for json_streamed_response in response:
            yield response_cls(**json_streamed_response)

    def _no_stream_response(self, response: Iterator[Dict[str, Any]], response_cls: Type[BaseModel]) -> BaseModel:
        for resp in response:
            return response_cls(**resp)
        raise ShuttleAIException("No response received")

    def handle_request(
        self,
        method: str,
        endpoint: str,
        request_data: Dict[str, Any] | None,
        response_cls: Type[BaseModel],
        stream: bool = False,
    ) -> Any:
        # assert issubclass(response_cls, BaseModel)
        response = self._client._request(  # type: ignore
            method=method,
            json=request_data,
            path=endpoint,
            stream=stream,
        )
        if stream:
            return self._stream_response(response, response_cls)
        else:
            return self._no_stream_response(response, response_cls)


class AsyncResource(BaseResource):
    async def _stream_response(
        self, response: AsyncIterator[Dict[str, Any]], response_cls: Type[BaseModel]
    ) -> AsyncIterable[BaseModel]:
        async for json_streamed_response in response:
            yield response_cls(**json_streamed_response)

    async def _no_stream_response(
        self, response: AsyncIterator[Dict[str, Any]], response_cls: Type[BaseModel]
    ) -> BaseModel:
        async for resp in response:
            return response_cls(**resp)
        raise ShuttleAIException("No response received")

    async def handle_request(
        self,
        method: str,
        endpoint: str,
        request_data: Dict[str, Any] | None,
        response_cls: Type[BaseModel],
        stream: bool = False,
    ) -> Any:
        assert issubclass(response_cls, BaseModel)
        response = self._client._request(  # type: ignore
            method=method,
            json=request_data,
            path=endpoint,
            stream=stream,
        )
        if stream:
            return self._stream_response(response, response_cls)
        else:
            return await self._no_stream_response(response, response_cls)



from functools import cached_property
from typing import Generic, Optional, Type, TypeVar

from shuttleai.client.base import ClientBase
from shuttleai.resources.common import AsyncResource, SyncResource, T
from shuttleai.schemas.images.generations import ImagesGenerationResponse


class AsyncGenerations(AsyncResource):
    async def generate(
        self,
        prompt: str,
        model: Optional[str] = None,
    ) -> ImagesGenerationResponse:
        request = self._client._make_image_request(
            prompt,
            model,
        )

        return await self.handle_request(  # type: ignore
            method="post",
            endpoint="v1/images/generations",
            request_data=request,
            response_cls=ImagesGenerationResponse,
        )


class SyncGenerations(SyncResource):
    def generate(
        self,
        prompt: str,
        model: Optional[str] = None,
    ) -> ImagesGenerationResponse:
        request = self._client._make_image_request(
            prompt,
            model,
        )

        return self.handle_request(  # type: ignore
            method="post",
            endpoint="v1/images/generations",
            request_data=request,
            response_cls=ImagesGenerationResponse,
        )


GenerationsType = TypeVar("GenerationsType", SyncGenerations, AsyncGenerations)


class BaseImages(Generic[T, GenerationsType]):
    _client: T
    _generations_class: Type[GenerationsType]
    _generations: Optional[GenerationsType]

    def __init__(self, client: T, generations_class: Type[GenerationsType]) -> None:
        self._client = client
        self._generations_class = generations_class
        self._generations = None

    @cached_property
    def generations(self) -> GenerationsType:
        if self._generations is None:
            self._generations = self._generations_class(self._client)
        return self._generations


class Images(BaseImages[ClientBase, SyncGenerations]):
    def __init__(self, client: ClientBase) -> None:
        super().__init__(client, SyncGenerations)


class AsyncImages(BaseImages[ClientBase, AsyncGenerations]):
    def __init__(self, client: ClientBase) -> None:
        super().__init__(client, AsyncGenerations)



from .audio.audio import AsyncAudio, Audio
from .chat.completions import AsyncChat, Chat
from .embeddings import AsyncEmbeddings, Embeddings
from .etc.insults import AsyncInsults, Insults
from .etc.jokes import AsyncJokes, Jokes
from .images.generations import AsyncImages, Images
from .moderations import AsyncModerations, Moderations

__all__ = [
    "AsyncChat",
    "Chat",
    "AsyncImages",
    "Images",
    "AsyncAudio",
    "Audio",
    "AsyncModerations",
    "Moderations",
    "AsyncEmbeddings",
    "Embeddings",
    "AsyncInsults",
    "Insults",
    "AsyncJokes",
    "Jokes",
    # Add new resources here
]



from typing import Final, Union

from aiohttp import ClientTimeout
from httpx import Timeout

DEFAULT_TIMEOUT: Final[float] = 2 * 60
DEFAULT_AIOTTP_TIMEOUT: Final[ClientTimeout] = ClientTimeout(total=DEFAULT_TIMEOUT)
DEFAULT_HTTPX_TIMEOUT: Final[Timeout] = Timeout(DEFAULT_TIMEOUT)

HTTPXTimeoutTypes = Union[
    float,
    "Timeout",
]

AIOHTTPTimeoutTypes = Union[float, "ClientTimeout"]

TimeoutTypes = Union[HTTPXTimeoutTypes, AIOHTTPTimeoutTypes]



from typing import Literal, Optional, Union

from shuttleai.resources.common import AsyncResource, SyncResource
from shuttleai.schemas.embeddings import EmbeddingResponse


class AsyncEmbeddings(AsyncResource):
    async def create(
        self,
        input: str,
        model: Optional[
            Union[str, Literal["text-embedding-3-small", "text-embedding-3-large"]]
        ] = "text-embedding-3-large",
    ) -> EmbeddingResponse:
        request = {"input": input, "model": model}

        return await self.handle_request(  # type: ignore
            method="post",
            endpoint="v1/embeddings",
            request_data=request,
            response_cls=EmbeddingResponse,
        )


class Embeddings(SyncResource):
    def create(
        self,
        input: str,
        model: Optional[
            Union[str, Literal["text-embedding-3-small", "text-embedding-3-large"]]
        ] = "text-embedding-3-large",
    ) -> EmbeddingResponse:
        request = {"input": input, "model": model}

        return self.handle_request(  # type: ignore
            method="post",
            endpoint="v1/embeddings",
            request_data=request,
            response_cls=EmbeddingResponse,
        )



from pydantic import BaseModel, Field


class ModerationCategoryScores(BaseModel):
    sexual: float = Field(alias="sexual")
    hate: float = Field(alias="hate")
    harassment: float = Field(alias="harassment")
    self_harm: float = Field(alias="self-harm")
    sexual_minors: float = Field(alias="sexual/minors")
    hate_threatening: float = Field(alias="hate/threatening")
    violence_graphic: float = Field(alias="violence/graphic")
    self_harm_intent: float = Field(alias="self-harm/intent")
    self_harm_instructions: float = Field(alias="self-harm/instructions")
    harassment_threatening: float = Field(alias="harassment/threatening")
    violence: float = Field(alias="violence")


class ModerationResult(BaseModel):
    flagged: bool
    categories: ModerationCategoryScores
    category_scores: ModerationCategoryScores


class ModerationResponse(BaseModel):
    id: str
    model: str
    results: list[ModerationResult]

    def is_flagged(self) -> bool:
        return any(result.flagged for result in self.results)



from __future__ import annotations

from typing import Dict, Optional

from aiohttp import ClientResponse
from httpx import Response


class ShuttleAIException(Exception):
    """Base Exception class, returned when nothing more specific applies"""

    def __init__(self, message: Optional[str] = None) -> None:
        super(ShuttleAIException, self).__init__(message)

        self.message = message

    def __str__(self) -> str:
        msg = self.message or "<null>"
        return msg

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(message={str(self)})"


class ShuttleAIAPIException(ShuttleAIException):
    """Returned when the API responds with an error message"""

    def __init__(
        self,
        message: Optional[str] = None,
        http_status: Optional[int] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> None:
        super().__init__(message)
        self.http_status = http_status
        self.headers = headers or {}

    @classmethod
    def from_response(cls, response: Response | ClientResponse, message: Optional[str] = None) -> ShuttleAIAPIException:
        return cls(
            message=(message or response.text if isinstance(response, Response) else response.reason),
            http_status=(response.status_code if isinstance(response, Response) else response.status),
            headers=dict(response.headers) if isinstance(response, Response) else dict(response.headers.items()),
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(message={str(self)}, http_status={self.http_status}, headers={self.headers})"


class ShuttleAIAPIStatusException(ShuttleAIAPIException):
    """Returned when we receive a non-200 response from the API that we should retry"""


class ShuttleAIConnectionException(ShuttleAIException):
    """Returned when the SDK can not reach the API server for any reason"""



from typing import Optional

from pydantic import BaseModel


class SupportsTools(BaseModel):
    regular: bool
    """Whether the model supports regular tool calling."""

    streamed: bool
    """Whether the model supports streamed tool calling."""

    parallel: bool
    """Whether the model supports parallel tool calling."""


class SupportsMaxTokens(BaseModel):
    input: int
    """The maximum number of tokens that can be input to the model."""

    output: int
    """The maximum number of tokens that can be output from the model."""


class Capabilities(BaseModel):
    supports_tools: Optional[SupportsTools] = None
    """The tool calling capabilities of the model."""

    supports_max_tokens: Optional[SupportsMaxTokens] = None
    """The maximum token capabilities of the model."""

    supports_image_input: bool
    """Whether the model supports image input."""

    supports_audio_input: bool
    """Whether the model supports audio input."""



__title__ = "shuttleai"
__version__ = "4.5.6"

import json
import time
import typing
from pathlib import Path

import requests
from packaging import version

from ._patch import _patch_httpx
from .client import AsyncShuttleAI, ShuttleAI

CACHE_FILE = Path(f"{__title__}-version.json")
CACHE_DURATION = 86400  # 24 hours in seconds


def read_cached_version_info() -> typing.Optional[dict[str, typing.Any]]:
    """Read the cached version information from the cache file."""
    if CACHE_FILE.exists():
        try:
            with open(CACHE_FILE, "r") as file:
                return json.load(file)  # type: ignore
        except (json.JSONDecodeError, IOError) as e:
            print(f"Error reading cache file: {e}")
    return None


def write_cached_version_info(version_info: dict[str, typing.Any]) -> None:
    """Write the version information to the cache file."""
    try:
        with open(CACHE_FILE, "w") as file:
            json.dump(version_info, file)
    except IOError as e:
        print(f"Error writing to cache file: {e}")


def is_cache_valid(cache_time: float) -> bool:
    """Check if the cache is still valid based on the cache duration."""
    return time.time() - cache_time < CACHE_DURATION


def check_for_updates() -> None:
    """Check for updates and notify the user if a newer version is available."""
    cached_version_info = read_cached_version_info()

    if cached_version_info:
        cache_time = cached_version_info.get("time")
        if cache_time and is_cache_valid(cache_time):
            latest_version = cached_version_info.get("version")
            if latest_version and version.parse(__version__) < version.parse(latest_version):
                print_update_message(latest_version)
            return

    try:
        response = requests.get("https://pypi.org/pypi/shuttleai/json")
        response.raise_for_status()
        latest_version = response.json()["info"]["version"]
        write_cached_version_info({"version": latest_version, "time": time.time()})
        if version.parse(__version__) < version.parse(latest_version):
            print_update_message(latest_version)
    except requests.RequestException as e:
        print(f"Could not check for updates: {e}")


def print_update_message(latest_version: str) -> None:
    """Print a message to the user indicating that an update is available."""
    print(
        f"WARNING: You are using an outdated version of {__title__} ({__version__}). "
        f"The latest version is {latest_version}. It is recommended to upgrade using:\n"
        f">> pip install -U {__title__}"
    )


_patch_httpx()
check_for_updates()


__all__ = ["ShuttleAI", "AsyncShuttleAI"]



import logging
import os
from abc import ABC
from typing import Any, Dict, List, Optional, Union

import orjson

from shuttleai import __version__
from shuttleai._types import TimeoutTypes
from shuttleai.exceptions import ShuttleAIException
from shuttleai.schemas.chat.completions import ChatMessage, Function, ToolChoice


class ClientBase(ABC):  # noqa: B024
    _timeout: TimeoutTypes
    _api_key: Optional[str]
    _base_url: Optional[str]
    _logger: logging.Logger
    _default_chat_model: str
    _default_image_model: str
    _default_audio_speech_model: str
    _version: str

    # client options
    api_key: str
    base_url: str

    def __init__(
        self,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        timeout: TimeoutTypes = 120.0,
    ):
        self._timeout = timeout
        self._api_key = api_key or os.getenv("SHUTTLEAI_API_KEY")
        if not self._api_key:
            raise ShuttleAIException("API key not provided. Please set SHUTTLEAI_API_KEY environment variable.")

        self._base_url = base_url or os.getenv("SHUTTLEAI_API_BASE")
        if not self._base_url:
            self._base_url = "https://api.shuttleai.app"
        # else: # TODO: uh, yeah, we might just follow the OpenAI SDK format, but for now, let's just keep as is
        #     self._base_url = self._base_url.rstrip("/v1")

        self.api_key = self._api_key
        self.base_url = self._base_url

        self._logger = logging.getLogger(__name__)
        self._default_chat_model = "shuttle-2.5"
        self._default_image_model = "shuttle-2-diffusion"
        self._default_audio_speech_model = "eleven_turbo_v2_5"
        self._version = __version__

        if "shuttleai.app" not in self.base_url:
            if "api.openai.com" not in self.base_url:
                self._logger.warning(
                    "You are using an **unofficial, unverified** non-ShuttleAI URL. \
                    This is not recommended and may lead to malfunctions. \
                    Your data could be at risk since you are using a 3rd party. \
                    Please use the official ShuttleAI API URL: https://api.shuttleai.app/v1"
                )
            else:
                self._logger.warning(
                    "You are using the official, verified OpenAI API URL. \
                    This library is not meant to replace the OpenAI SDK. \
                    If you wish to use the OpenAI API, consider using their SDK respectively. \
                    Otherwise, please use the official ShuttleAI API URL: https://api.shuttleai.app/v1"
                )
            self._default_chat_model = "gpt-4o-mini"
            self._default_image_model = "dall-e-3"
            self._default_audio_speech_model = "whisper-1"

        self._logger.info(f"ShuttleAI API client initialized with base URL: {self._base_url}")

    def _build_sampling_params(
        self,
        max_tokens: Optional[int],
        temperature: Optional[float],
        top_p: Optional[float],
    ) -> Dict[str, Any]:
        return {
            k: v
            for k, v in {
                "temperature": temperature,
                "max_tokens": max_tokens,
                "top_p": top_p,
            }.items()
            if v is not None
        }

    def _parse_tools(self, tools: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        return [
            {
                "type": tool["type"],
                "function": (
                    tool["function"].model_dump(exclude_none=True)
                    if isinstance(tool["function"], Function)
                    else tool["function"]
                ),
            }
            for tool in tools
            if tool["type"] == "function"
        ]

    def _parse_tool_choice(self, tool_choice: Union[str, ToolChoice]) -> str:
        return tool_choice.value if isinstance(tool_choice, ToolChoice) else tool_choice

    def _parse_messages(self, messages: List[Any]) -> List[Dict[str, Any]]:
        return [
            (message.model_dump(exclude_none=True) if isinstance(message, ChatMessage) else message)
            for message in messages
        ]

    def _make_request(self, endpoint: str, request_data: Dict[str, Any]) -> Dict[str, Any]:
        if "model" not in request_data:
            request_data["model"] = getattr(self, f"_default_{endpoint}_model")
        self._logger.debug(f"{endpoint.capitalize()} request: {request_data}")
        return request_data

    def _make_chat_request(
        self,
        messages: List[Any],
        model: Optional[str] = None,
        image: Optional[str] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
        top_p: Optional[float] = None,
        stream: Optional[bool] = None,
        tool_choice: Optional[Union[str, ToolChoice]] = None,
    ) -> Dict[str, Any]:
        request_data: Dict[str, Any] = {
            "messages": self._parse_messages(messages),
        }
        if model:
            request_data["model"] = model
        if image:
            request_data["image"] = image
        if tools:
            request_data["tools"] = self._parse_tools(tools)
        if tool_choice:
            request_data["tool_choice"] = self._parse_tool_choice(tool_choice)
        if stream:
            request_data["stream"] = stream
        request_data.update(self._build_sampling_params(max_tokens, temperature, top_p))
        return self._make_request("chat", request_data)

    def _make_image_request(self, prompt: str, model: Optional[str] = None) -> Dict[str, Any]:
        request_data: Dict[str, Any] = {
            "prompt": prompt,
        }
        if model:
            request_data["model"] = model
        return self._make_request("image", request_data)

    def _make_audio_speech_request(
        self,
        input: str,
        model: str = "eleven_turbo_v2_5",
        voice: Optional[str] = None,
    ) -> Dict[str, Any]:
        request_data: Dict[str, Any] = {
            "input": input,
            "model": model,
            **({"voice": voice} if voice else {}),
        }
        return self._make_request("audio_speech", request_data)

    def _make_audio_trans_request(  # translations/transcriptions share similar request/response schemas
        self,
        file: str,
        model: Optional[str] = None,
    ) -> Dict[str, Any]:
        request_data: Dict[str, Any] = {
            "file": file,
        }
        if model:
            request_data["model"] = model
        return self._make_request("audio_trans", request_data)

    def _process_line(self, line: Union[str, bytes]) -> Optional[Dict[str, Any] | Any]:
        line = line.encode("utf-8") if isinstance(line, str) else line
        if line.startswith(b"data: "):
            line = line[6:].strip()
            if line != b"[DONE]":
                return orjson.loads(line)
        return None



import argparse
import json
import logging
import os
import readline
import sys
from typing import Dict, KeysView, List, Optional, Union

import yaml

from shuttleai import ShuttleAI
from shuttleai.schemas.chat.completions import ChatMessage

MODEL_LIST: List[str] = [
    "shuttleai/shuttle-2.5",
    "shuttle-2-turbo",
    "shuttle-2.5",
    "shuttleai/shuttle-2.5-mini",
    "shuttle-2.5-mini",
    "openai/gpt-4o-mini-2024-07-18",
    "gpt-4o-mini",
    "gpt-4o-mini-2024-07-18",
    "openai/chatgpt-4o-latest",
    "chatgpt-4o",
    "chatgpt-4o-latest",
    "openai/gpt-4o-2024-08-06",
    "gpt-4o-16k",
    "gpt-4o-2024-08-06",
    "openai/gpt-4o-2024-05-13",
    "gpt-4o",
    "gpt-4o-2024-05-13",
    "openai/gpt-4-turbo-2024-04-09",
    "gpt-4-turbo",
    "gpt-4-turbo-2024-04-09",
    "openai/gpt-4-0125-preview",
    "gpt-4-turbo-preview",
    "gpt-4-0125-preview",
    "openai/gpt-4-1106-preview",
    "gpt-4-1106-preview",
    "openai/gpt-4-0613",
    "gpt-4",
    "gpt-4-0613",
    "openai/gpt-3.5-turbo-0125",
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-16k",
    "gpt-3.5-turbo-0125",
    "openai/gpt-3.5-turbo-1106",
    "gpt-3.5-turbo-1106",
    "anthropic/claude-3-5-sonnet-20240620",
    "claude-3-5-sonnet",
    "claude-3.5-sonnet",
    "claude-3-5-sonnet-20240620",
    "anthropic/claude-3-haiku-20240307",
    "claude-3-haiku",
    "claude-3-haiku-20240307",
    "meta-llama/meta-llama-3.1-405b-instruct",
    "llama-3.1-405b",
    "meta-llama-3.1-405b-instruct",
    "meta-llama/meta-llama-3.1-70b-instruct",
    "llama-3.1-70b",
    "meta-llama-3.1-70b-instruct",
    "meta-llama/meta-llama-3.1-8b-instruct",
    "llama-3.1-8b",
    "meta-llama-3.1-8b-instruct",
]
DEFAULT_MODEL: str = "shuttle-2.5"
LOG_FORMAT: str = "%(asctime)s - %(levelname)s - %(message)s"
# A dictionary of all commands and their arguments, used for tab completion.
COMMAND_LIST: Dict[str, Union[Dict[str, Dict], Dict]] = {
    "/new": {},
    "/help": {},
    "/model": {model: {} for model in MODEL_LIST},  # Nested completions for models
    "/system": {},
    "/config": {},
    "/download": {"json": {}, "txt": {}, "yaml": {}},
    "/quit": {},
    "/exit": {},
}

logger = logging.getLogger("chatbot")


def find_completions(
    command_dict: Dict[str, Union[Dict[str, Dict], Dict]], parts: list[str]
) -> Union[KeysView[str], List[str]]:
    if not parts:
        return command_dict.keys()
    if parts[0] in command_dict:
        return find_completions(command_dict[parts[0]], parts[1:])
    else:
        return [cmd for cmd in command_dict if cmd.startswith(parts[0])]


def completer(text: str, state: int) -> Optional[str]:
    buffer = readline.get_line_buffer()  # type: ignore
    line_parts = buffer.lstrip().split(" ")
    options = find_completions(COMMAND_LIST, line_parts[:-1])

    try:
        return [option for option in options if option.startswith(line_parts[-1])][state]
    except IndexError:
        return None


readline.set_completer(completer)  # type: ignore
readline.set_completer_delims(" ")  # type: ignore
# Enable tab completion
readline.parse_and_bind("tab: complete")  # type: ignore


class ChatBot:
    def __init__(self, api_key: str, model: str, system_message: Optional[str] = None):
        if not api_key:
            raise ValueError("An API key must be provided to use the ShuttleAI API.")
        self.client = ShuttleAI(api_key=api_key)
        self.model = model
        self.system_message = system_message
        self.messages: List[ChatMessage] = []

    def opening_instructions(self) -> None:
        print(
            """
To chat: type your message and hit enter
To start a new chat: /new
To switch model: /model <model name>
To switch system message: /system <message>
To see current config: /config
To exit: /exit, /quit, or hit CTRL+C
To see this help: /help
HINT: We support TAB autocompletion for commands and model names!
"""
        )

    def new_chat(self) -> None:
        print("")
        print(f"Starting new chat with model: \033[38;5;105m{self.model}\033[0m")
        print("")
        self.messages = []
        if self.system_message:
            self.messages.append(ChatMessage(role="system", content=self.system_message))

    def switch_model(self, input: str) -> None:
        model = self.get_arguments(input)
        if model in MODEL_LIST:
            self.model = model
            logger.info(f"Switched to model \033[38;5;105m{model}\033[")
        else:
            logger.error(f"Invalid model name: {model}")

    def switch_system_message(self, input: str) -> None:
        system_message = self.get_arguments(input)
        if system_message:
            self.system_message = system_message
            logger.info(f"Switching system message: {system_message}")
            self.new_chat()
        else:
            logger.error(f"Invalid system message: {system_message}")

    def show_config(self) -> None:
        print("")
        print(f"Current model: \033[38;5;105m{self.model}\033[0m")
        print(f"Current system message: {self.system_message}")
        print("")

    def collect_user_input(self) -> str:
        print("")
        return input("\033[38;2;50;168;82mUser: \033[0m")

    def run_inference(self, content: str) -> None:
        print("")
        print("\033[38;5;105mSHUTTLEAI\033[0m:")
        print("")

        self.messages.append(ChatMessage(role="user", content=content))

        assistant_response = ""
        logger.debug(f"Running inference with model: {self.model}")
        logger.debug(f"Sending messages: {self.messages}")
        for chunk in self.client.chat.completions.create(
            model=self.model,
            messages=self.messages,
            stream=True
        ):
            if response := chunk.first_choice.delta.content:
                print(response, end="", flush=True)
                assistant_response += response

        print("", flush=True)

        if assistant_response:
            self.messages.append(ChatMessage(role="assistant", content=assistant_response))
        logger.debug(f"Current messages: {self.messages}")

    def get_command(self, input: str) -> str:
        return input.split()[0].strip()

    def get_arguments(self, input: str) -> str:
        try:
            return " ".join(input.split()[1:])
        except IndexError:
            return ""

    def is_command(self, input: str) -> bool:
        return self.get_command(input) in COMMAND_LIST

    def execute_command(self, input: str) -> None:
        command = self.get_command(input)
        if command in ["/exit", "/quit"]:
            self.exit()
        elif command == "/help":
            self.opening_instructions()
        elif command == "/new":
            self.new_chat()
        elif command == "/model":
            self.switch_model(input)
        elif command == "/system":
            self.switch_system_message(input)
        elif command == "/config":
            self.show_config()
        elif command == "/download":
            self.download_conversation(input)

    def download_conversation(self, input: str) -> None:
        format = self.get_arguments(input).lower()
        if format not in ["json", "txt", "yaml"]:
            logger.error("Invalid format. Choose from 'json', 'txt', or 'yaml'.")
            return

        filename = f"conversation.{format}"
        try:
            if format == "json":
                with open(filename, "w") as f:
                    json.dump([msg.model_dump(exclude_none=True) for msg in self.messages], f, indent=4)
            elif format == "txt":
                with open(filename, "w") as f:
                    for msg in self.messages:
                        f.write(f"{msg.role.capitalize()}: {msg.content}\n\n")
            elif format == "yaml":
                with open(filename, "w") as f:
                    yaml.dump([msg.model_dump(exclude_none=True) for msg in self.messages], f, default_flow_style=False)
            logger.info(f"Conversation saved to {filename}")
        except Exception as e:
            logger.error(f"Error saving conversation: {e}")

    def start(self) -> None:
        self.opening_instructions()
        self.new_chat()
        while True:
            try:
                input = self.collect_user_input()
                if self.is_command(input):
                    self.execute_command(input)
                else:
                    self.run_inference(input)

            except KeyboardInterrupt:
                self.exit()

    def exit(self) -> None:
        logger.debug("Exiting chatbot")
        sys.exit(0)


def main() -> None:
    parser = argparse.ArgumentParser(description="A simple chatbot using the ShuttleAI API")
    parser.add_argument(
        "--api-key",
        default=os.environ.get("SHUTTLEAI_API_KEY"),
        help="ShuttleAI API key. Defaults to environment variable SHUTTLEAI_API_KEY",
    )
    parser.add_argument(
        "-m",
        "--model",
        choices=MODEL_LIST,
        default=DEFAULT_MODEL,
        help="Model for chat inference. Choices are %(choices)s. Defaults to %(default)s",
    )
    parser.add_argument("-s", "--system-message", help="Optional system message to prepend.")
    parser.add_argument("-d", "--debug", action="store_true", help="Enable debug logging")

    args = parser.parse_args()

    if args.debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    formatter = logging.Formatter(LOG_FORMAT)

    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    logger.debug(f"Starting chatbot with model: {args.model}, " f"system message: {args.system_message}")

    try:
        bot = ChatBot(args.api_key, args.model, args.system_message)
        bot.start()
    except Exception as e:
        logger.error(e)
        sys.exit(1)


if __name__ == "__main__":
    main()



import posixpath
from typing import Any, AsyncIterable, AsyncIterator, Dict, Literal, Mapping, Optional, Type, Union, overload

import aiohttp
import orjson
import pydantic_core
from aiofiles import open as aopen
from aiohttp import ClientTimeout

from shuttleai import resources
from shuttleai._types import DEFAULT_AIOTTP_TIMEOUT, AIOHTTPTimeoutTypes
from shuttleai.client.base import ClientBase
from shuttleai.exceptions import (
    ShuttleAIAPIException,
    ShuttleAIAPIStatusException,
    ShuttleAIConnectionException,
    ShuttleAIException,
)
from shuttleai.schemas.chat.completions import ChatCompletionResponse, ChatCompletionStreamResponse
from shuttleai.schemas.models.models import (
    BaseModelCard,
    ListModelsResponse,
    ListVerboseModelsResponse,
    ProxyCard,
)


class AsyncShuttleAI(ClientBase):
    """
    Asynchronous wrapper for the ShuttleAI API
    """

    default_headers: Mapping[str, str] | None = None

    def __init__(
        self,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        timeout: AIOHTTPTimeoutTypes = DEFAULT_AIOTTP_TIMEOUT,
        default_headers: Mapping[str, str] | None = None,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        super().__init__(base_url, api_key, timeout)

        if self.api_key is None:
            raise ShuttleAIException(
                "API key not provided. Please set SHUTTLEAI_API_KEY environment variable.\n"
                + "Alternatively, you may pass it as an argument to the ShuttleAI class constructor as `api_key`.\n"
                + "In addition to that, you can also set the api key after creating the ShuttleAI object by setting \
                    `client.api_key`.\n"
            )

        self._timeout = timeout if isinstance(timeout, ClientTimeout) else ClientTimeout(total=timeout)
        if default_headers:
            self.default_headers = default_headers

        self._session: Optional[aiohttp.ClientSession] = None
        if session:
            self._session = session

        self.chat: resources.AsyncChat = resources.AsyncChat(self)
        self.images: resources.AsyncImages = resources.AsyncImages(self)
        self.audio: resources.AsyncAudio = resources.AsyncAudio(self)
        self.moderations: resources.AsyncModerations = resources.AsyncModerations(self)
        self.embeddings: resources.AsyncEmbeddings = resources.AsyncEmbeddings(self)
        self.insults: resources.AsyncInsults = resources.AsyncInsults(self)
        self.jokes: resources.AsyncJokes = resources.AsyncJokes(self)

    async def __aenter__(self) -> "AsyncShuttleAI":
        if self._session is None:
            self._session = aiohttp.ClientSession(timeout=self._timeout)
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[Type[BaseException]],
    ) -> None:
        await self.close()

    async def close(self) -> None:
        if self._session:
            await self._session.close()
            self._session = None

    async def _check_response_status_codes(self, response: aiohttp.ClientResponse) -> None:
        if response.status in {429, 500, 502, 503, 504}:
            raise ShuttleAIAPIStatusException.from_response(
                response,
                message=f"Status: {response.status}. Message: {await response.text()}",
            )
        elif 400 <= response.status < 500:
            raise ShuttleAIAPIException.from_response(
                response,
                message=f"Status: {response.status}. Message: {await response.text()}",
            )
        elif response.status >= 500:
            raise ShuttleAIException(
                message=f"Status: {response.status}. Message: {await response.text()}",
            )

    async def _check_response(self, response: aiohttp.ClientResponse) -> Dict[str, Any]:
        await self._check_response_status_codes(response)

        json_response: Dict[str, Any] = orjson.loads(await response.read())

        return json_response

    async def _request(
        self,
        method: str,
        json: Optional[Dict[str, Any]],
        path: str,
        stream: bool = False,
    ) -> AsyncIterator[Dict[str, Any]]:
        if self._session is None:
            self._session = aiohttp.ClientSession(timeout=self._timeout)

        json_bytes: Any | None = None
        if json and len(json) > 0:
            if "file" in json:
                async with aopen(json["file"], "rb") as f:
                    json_bytes = {"file": await f.read()}
            else:
                json_bytes = orjson.dumps(json)

        # json_bytes: bytes | None = (
        #     orjson.dumps(json) if json and len(json) > 0 else None
        # )  # x-sai [dict to bytes]

        accept_header = "text/event-stream" if stream else "application/json"
        headers = {
            "Accept": accept_header,
            "User-Agent": f"shuttleai-python/a-{self._version}",
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }

        if self.default_headers:
            headers.update(self.default_headers)

        url = posixpath.join(self.base_url, path)

        self._logger.debug(f"Sending request: {method} {url} {json}")

        try:
            async with self._session.request(
                method,
                url,
                headers=headers,
                data=json_bytes,
            ) as response:
                if stream:
                    async for line in response.content:
                        json_streamed_response = self._process_line(line)
                        if json_streamed_response:
                            yield json_streamed_response
                else:
                    yield await self._check_response(response)

        except aiohttp.ClientConnectorError as e:
            raise ShuttleAIConnectionException(str(e)) from e
        except aiohttp.ClientError as e:
            raise ShuttleAIException(f"Unexpected exception ({e.__class__.__name__}): {e}") from e
        except orjson.JSONDecodeError as e:
            raise ShuttleAIAPIException.from_response(
                response,
                message=f"Failed to decode json body: {await response.text()}",
            ) from e
        except ShuttleAIAPIStatusException as e:
            raise ShuttleAIAPIStatusException.from_response(response, message=str(e)) from e

    async def fetch_model(self, model_id: str) -> BaseModelCard:
        """Fetches a model by its ID

        Args:
            model_id (str): The ID of the model to fetch

        Returns:
            BaseModelCard, None]: The model if it exists
        """
        singleton_response = self._request("get", {}, f"v1/models/{model_id}")
        try:
            return BaseModelCard(**(await singleton_response.__anext__())["data"])
        except (pydantic_core.ValidationError, StopAsyncIteration) as e:
            raise ShuttleAIException("No response received") from e

    async def list_models(self) -> Union[ListModelsResponse, ListVerboseModelsResponse]:
        """Returns a list of the available models

        Returns:
            ListModelsResponse: A response object containing the list of models.
        """
        return await self._fetch_and_process_models("v1/models", ListModelsResponse)

    async def list_models_verbose(
        self,
    ) -> Union[ListVerboseModelsResponse, ListModelsResponse]:
        """Returns a list of the available models with verbose information

        Returns:
            ListVerboseModelsResponse: A response object containing the list of models.
        """
        return await self._fetch_and_process_models("v1/models/verbose", ListVerboseModelsResponse)

    async def _fetch_and_process_models(
        self,
        endpoint: str,
        response_class: Type[Union[ListModelsResponse, ListVerboseModelsResponse]],
    ) -> Union[ListModelsResponse, ListVerboseModelsResponse]:
        singleton_response = self._request("get", {}, endpoint)
        try:
            list_models_response = response_class(**(await singleton_response.__anext__()))
        except pydantic_core.ValidationError as e:
            raise ShuttleAIException("No response received") from e

        models_by_id = {model.id: model for model in list_models_response.data}

        for model in list_models_response.data:
            if isinstance(model, ProxyCard):
                model_parent = models_by_id.get(model.proxy_to)
                assert isinstance(model_parent, BaseModelCard)
                if model_parent:
                    model.parent = model_parent

        return list_models_response

    @overload
    async def ez_chat(  # type: ignore
        self, text: str, model: Optional[str] = None, stream: Literal[False] = False
    ) -> ChatCompletionResponse: ...

    @overload
    async def ez_chat(
        self, text: str, model: Optional[str] = None, stream: Literal[True] = True
    ) -> AsyncIterable[ChatCompletionStreamResponse]: ...

    async def ez_chat(
        self,
        text: str,
        model: Optional[str] = None,
        stream: bool = False,
    ) -> Union[ChatCompletionResponse, AsyncIterable[ChatCompletionStreamResponse]]:
        return await self.chat.completions.create(  # type: ignore
            messages=[{"role": "user", "content": text}],
            model=model,
            stream=stream,
        )



from functools import cached_property
from typing import (
    Any,
    AsyncIterable,
    Dict,
    Generic,
    Iterable,
    List,
    Literal,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)

from shuttleai.client.base import ClientBase
from shuttleai.resources.common import AsyncResource, SyncResource, T
from shuttleai.schemas.chat.completions import (
    ChatCompletionResponse,
    ChatCompletionStreamResponse,
    ChatMessage,
    ToolChoice,
)


class AsyncCompletions(AsyncResource):
    @overload
    async def create(  # type: ignore
        self,
        messages: Union[List[ChatMessage], List[Dict[str, Any]]],
        model: Optional[str] = None,
        image: Optional[str] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
        top_p: Optional[float] = None,
        tool_choice: Optional[Union[str, ToolChoice]] = None,
        stream: Literal[False] = False,
    ) -> ChatCompletionResponse: ...

    @overload
    async def create(
        self,
        messages: Union[List[ChatMessage], List[Dict[str, Any]]],
        model: Optional[str] = None,
        image: Optional[str] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
        top_p: Optional[float] = None,
        tool_choice: Optional[Union[str, ToolChoice]] = None,
        stream: Literal[True] = True,
    ) -> AsyncIterable[ChatCompletionStreamResponse]: ...

    async def create(
        self,
        messages: Union[List[ChatMessage], List[Dict[str, Any]]],
        model: Optional[str] = None,
        image: Optional[str] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
        top_p: Optional[float] = None,
        tool_choice: Optional[Union[str, ToolChoice]] = None,
        stream: bool = False,
    ) -> Union[ChatCompletionResponse, AsyncIterable[ChatCompletionStreamResponse]]:
        request = self._client._make_chat_request(
            messages,
            model,
            image=image,
            tools=tools,
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
            stream=stream,
            tool_choice=tool_choice,
        )

        return await self.handle_request(  # type: ignore
            method="post",
            endpoint="v1/chat/completions",
            request_data=request,
            response_cls=ChatCompletionStreamResponse if stream else ChatCompletionResponse,  # type: ignore
            stream=stream,
        )


class SyncCompletions(SyncResource):
    @overload
    def create(  # type: ignore
        self,
        messages: Union[List[ChatMessage], List[Dict[str, Any]]],
        model: Optional[str] = None,
        image: Optional[str] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
        top_p: Optional[float] = None,
        tool_choice: Optional[Union[str, ToolChoice]] = None,
        stream: Literal[False] = False,
    ) -> ChatCompletionResponse: ...

    @overload
    def create(
        self,
        messages: Union[List[ChatMessage], List[Dict[str, Any]]],
        model: Optional[str] = None,
        image: Optional[str] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
        top_p: Optional[float] = None,
        tool_choice: Optional[Union[str, ToolChoice]] = None,
        stream: Literal[True] = True,
    ) -> Iterable[ChatCompletionStreamResponse]: ...

    def create(
        self,
        messages: Union[List[ChatMessage], List[Dict[str, Any]]],
        model: Optional[str] = None,
        image: Optional[str] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
        top_p: Optional[float] = None,
        tool_choice: Optional[Union[str, ToolChoice]] = None,
        stream: bool = False,
    ) -> Union[ChatCompletionResponse, Iterable[ChatCompletionStreamResponse]]:
        request = self._client._make_chat_request(
            messages,
            model,
            image=image,
            tools=tools,
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
            stream=stream,
            tool_choice=tool_choice,
        )

        return self.handle_request(  # type: ignore
            method="post",
            endpoint="v1/chat/completions",
            request_data=request,
            response_cls=ChatCompletionStreamResponse if stream else ChatCompletionResponse,  # type: ignore
            stream=stream,
        )


CompletionsType = TypeVar("CompletionsType", SyncCompletions, AsyncCompletions)


class BaseChat(Generic[T, CompletionsType]):
    _client: T
    _completions_class: Type[CompletionsType]

    def __init__(self, client: T, completions_class: Type[CompletionsType]) -> None:
        self._client = client
        self._completions_class = completions_class

    @cached_property
    def completions(self) -> CompletionsType:
        return self._completions_class(self._client)


class Chat(BaseChat[ClientBase, SyncCompletions]):
    def __init__(self, client: ClientBase) -> None:
        super().__init__(client, SyncCompletions)


class AsyncChat(BaseChat[ClientBase, AsyncCompletions]):
    def __init__(self, client: ClientBase) -> None:
        super().__init__(client, AsyncCompletions)



from typing import Optional

from shuttleai.resources.common import AsyncResource, SyncResource
from shuttleai.schemas.audio.speech import AudioSpeechResponse


class AsyncSpeech(AsyncResource):
    async def generate(
        self,
        input: str,
        model: str = "eleven-labs",
        voice: Optional[str] = None,
    ) -> AudioSpeechResponse:
        request = self._client._make_audio_speech_request(input, model, voice)

        return await self.handle_request(  # type: ignore
            method="post",
            endpoint="v1/audio/speech",
            request_data=request,
            response_cls=AudioSpeechResponse,
        )


class SyncSpeech(SyncResource):
    def generate(
        self,
        input: str,
        model: str = "eleven-labs",
        voice: Optional[str] = None,
    ) -> AudioSpeechResponse:
        request = self._client._make_audio_speech_request(input, model, voice)

        return self.handle_request(  # type: ignore
            method="post",
            endpoint="v1/audio/speech",
            request_data=request,
            response_cls=AudioSpeechResponse,
        )



from functools import cached_property
from typing import Generic, Type, TypeVar

from shuttleai.client.base import ClientBase
from shuttleai.resources.audio.speech import AsyncSpeech, SyncSpeech
from shuttleai.resources.audio.transcriptions import (
    AsyncTranscriptions,
    SyncTranscriptions,
)
from shuttleai.resources.audio.translations import AsyncTranslations, SyncTranslations
from shuttleai.resources.common import T

SpeechType = TypeVar("SpeechType", SyncSpeech, AsyncSpeech)
TranscriptionsType = TypeVar("TranscriptionsType", SyncTranscriptions, AsyncTranscriptions)
TranslationsType = TypeVar("TranslationsType", SyncTranslations, AsyncTranslations)


class BaseAudio(Generic[T, SpeechType, TranscriptionsType, TranslationsType]):
    _client: T
    _speech_class: Type[SpeechType]
    _transcriptions_class: Type[TranscriptionsType]
    _translations_class: Type[TranslationsType]

    def __init__(
        self,
        client: T,
        speech_class: Type[SpeechType],
        transcriptions_class: Type[TranscriptionsType],
        translations_class: Type[TranslationsType],
    ) -> None:
        self._client = client
        self._speech_class = speech_class
        self._transcriptions_class = transcriptions_class
        self._translations_class = translations_class

    @cached_property
    def speech(self) -> SpeechType:
        return self._speech_class(self._client)

    @cached_property
    def transcriptions(self) -> TranscriptionsType:
        return self._transcriptions_class(self._client)

    @cached_property
    def translations(self) -> TranslationsType:
        return self._translations_class(self._client)


class Audio(BaseAudio[ClientBase, SyncSpeech, SyncTranscriptions, SyncTranslations]):
    def __init__(self, client: ClientBase) -> None:
        super().__init__(client, SyncSpeech, SyncTranscriptions, SyncTranslations)


class AsyncAudio(BaseAudio[ClientBase, AsyncSpeech, AsyncTranscriptions, AsyncTranslations]):
    def __init__(self, client: ClientBase) -> None:
        super().__init__(client, AsyncSpeech, AsyncTranscriptions, AsyncTranslations)



from typing import Optional

from shuttleai.resources.common import AsyncResource, SyncResource
from shuttleai.schemas.audio.transcriptions import AudioTranscriptionResponse


class AsyncTranscriptions(AsyncResource):
    async def create(
        self,
        file: str,
        model: Optional[str] = None,
    ) -> AudioTranscriptionResponse:
        request = self._client._make_audio_trans_request(
            file,
            model,
        )

        return await self.handle_request(  # type: ignore
            method="post",
            endpoint="v1/audio/transcriptions",
            request_data=request,
            response_cls=AudioTranscriptionResponse,
        )


class SyncTranscriptions(SyncResource):
    def create(
        self,
        file: str,
        model: Optional[str] = None,
    ) -> AudioTranscriptionResponse:
        request = self._client._make_audio_trans_request(
            file,
            model,
        )

        return self.handle_request(  # type: ignore
            method="post",
            endpoint="v1/audio/transcriptions",
            request_data=request,
            response_cls=AudioTranscriptionResponse,
        )



from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel

from shuttleai.exceptions import ShuttleAIException
from shuttleai.schemas.common import UsageInfo


class Function(BaseModel):
    name: str
    description: str
    parameters: dict


class ToolType(str, Enum):
    function = "function"


class FunctionCall(BaseModel):
    name: str
    arguments: str


class ContentPartType(str, Enum):
    text = "text"
    image_url = "image_url"


class ToolCall(BaseModel):
    id: str = "call_null"
    type: ToolType = ToolType.function
    function: FunctionCall


class ToolChoice(str, Enum):
    auto: str = "auto"
    none: str = "none"


class ChatMessageContentPartText(BaseModel):
    type: ContentPartType = ContentPartType.text
    text: str


class ChatMessageContentPartImage(BaseModel):
    type: ContentPartType = ContentPartType.image_url
    image_url: str


class ChatMessage(BaseModel):
    role: str
    content: Optional[Union[str, List[Union[ChatMessageContentPartText, ChatMessageContentPartImage]]]] = None
    name: Optional[str] = None
    tool_calls: Optional[List[ToolCall]] = None
    # tool_call_id: Optional[str] = None


class DeltaMessage(BaseModel):
    role: Optional[str] = None
    content: Optional[str] = None
    tool_calls: Optional[List[ToolCall]] = None


class FinishReason(str, Enum):
    stop = "stop"
    length = "length"
    tool_calls = "tool_calls"


class ChatCompletionResponseStreamChoice(BaseModel):
    index: int
    delta: DeltaMessage
    finish_reason: Optional[FinishReason]


class ChatCompletionStreamResponse(BaseModel):
    id: str
    model: str
    choices: List[ChatCompletionResponseStreamChoice]
    created: Optional[int] = None
    object: Optional[str] = None
    usage: Optional[UsageInfo] = None

    @property
    def first_choice(self) -> ChatCompletionResponseStreamChoice:
        return self.choices[0]

    def print_chunk(self) -> None:
        try:
            print(f"Request ID: {self.id}")
            print(f"Model: {self.model}")
            print(f"Created: {self.created}")
            print(f"Usage: {self.usage}")
            for choice in self.choices:
                print(f"Index: {choice.index}")
                print(f"Delta: {choice.delta}")
                print(f"Finish Reason: {choice.finish_reason}")
        except Exception as e:
            raise ShuttleAIException(f"Error printing response: {e}") from e


class ChatCompletionResponseChoice(BaseModel):
    index: int
    message: ChatMessage
    finish_reason: Optional[FinishReason]


# class ShuttleAIMeta(BaseModel):
#     id: str
#     """The ID of the request."""

#     p: str
#     """The ID of the provider that processed the request.

#     The provider ID is semi-reliable, meaning upon VPS restarts, provider IDs may change;
#     however, they are guaranteed to remain the same for the duration of the VPS uptime.
#     (This can be useful for debugging/reporting purposes.)"""


class ChatCompletionResponse(BaseModel):
    id: str
    object: str
    created: int
    model: str
    choices: List[ChatCompletionResponseChoice]
    usage: UsageInfo

    @property
    def cost(self) -> float:
        return self.usage.total_charged

    # x_sai: Annotated[
    #     ShuttleAIMeta,
    #     Field(
    #         alias="x-sai",
    #         alias_priority=1,
    #         examples=[{"id": "req_123abc", "p": "p_123abc"}],
    #     ),
    # ]

    # @property
    # def xsai(self) -> ShuttleAIMeta:
    #     return self.x_sai

    # @property
    # def meta(self) -> ShuttleAIMeta:
    #     return self.x_sai

    # @property
    # def provider(self) -> str:
    #     return self.x_sai.p

    # @property
    # def provider_id(self) -> str:
    #     return self.x_sai.p

    # @property
    # def request_id(self) -> str:
    #     return self.x_sai.id

    @property
    def first_choice(self) -> ChatCompletionResponseChoice:
        return self.choices[0]

    def print(self) -> None:
        try:
            # print(f"Request ID: {self.request_id}")
            # print(f"Provider ID: {self.provider_id}")
            print(f"Model: {self.model}")
            print(f"Created: {self.created}")
            print(f"Usage: {self.usage}")
            for choice in self.choices:
                print(f"Index: {choice.index}")
                print(f"Message: {choice.message}")
                print(f"Finish Reason: {choice.finish_reason}")
        except Exception as e:
            raise ShuttleAIException(f"Error printing response: {e}") from e



from typing import Literal, Optional, Union

from shuttleai.resources.common import AsyncResource, SyncResource
from shuttleai.schemas.moderations import ModerationResponse


class AsyncModerations(AsyncResource):
    async def create(
        self,
        input: str,
        model: Optional[
            Union[str, Literal["text-moderation-latest", "text-moderation-stable"]]
        ] = "text-moderation-stable",
    ) -> ModerationResponse:
        request = {"input": input, "model": model}

        return await self.handle_request(  # type: ignore
            method="post",
            endpoint="v1/moderations",
            request_data=request,
            response_cls=ModerationResponse,
        )


class Moderations(SyncResource):
    def create(
        self,
        input: str,
        model: Optional[
            Union[str, Literal["text-moderation-latest", "text-moderation-stable"]]
        ] = "text-moderation-stable",
    ) -> ModerationResponse:
        request = {"input": input, "model": model}

        return self.handle_request(  # type: ignore
            method="post",
            endpoint="v1/moderations",
            request_data=request,
            response_cls=ModerationResponse,
        )



import posixpath
from json import JSONDecodeError
from typing import Any, Dict, Iterable, Iterator, Literal, Mapping, Optional, Type, Union, overload

import orjson
import pydantic_core
from httpx import Client, ConnectError, RequestError, Response

from shuttleai import resources
from shuttleai._types import DEFAULT_HTTPX_TIMEOUT, HTTPXTimeoutTypes
from shuttleai.client.base import ClientBase
from shuttleai.exceptions import (
    ShuttleAIAPIException,
    ShuttleAIAPIStatusException,
    ShuttleAIConnectionException,
    ShuttleAIException,
)
from shuttleai.schemas.chat.completions import ChatCompletionResponse, ChatCompletionStreamResponse
from shuttleai.schemas.models.models import (
    BaseModelCard,
    ListModelsResponse,
    ListVerboseModelsResponse,
    ProxyCard,
)


class ShuttleAI(ClientBase):
    """
    Synchronous wrapper for the ShuttleAI API
    """

    default_headers: Mapping[str, str] | None = None

    def __init__(
        self,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        timeout: HTTPXTimeoutTypes = DEFAULT_HTTPX_TIMEOUT,
        default_headers: Mapping[str, str] | None = None,
        http_client: Optional[Client] = None,
    ):
        super().__init__(base_url, api_key, timeout)

        if self.api_key is None:
            raise ShuttleAIException(
                "API key not provided. Please set SHUTTLEAI_API_KEY environment variable.\n"
                + "Alternatively, you may pass it as an argument to the ShuttleAI class constructor as `api_key`.\n"
                + "In addition to that, you can also set the api key after creating the ShuttleAI object by setting \
                    `client.api_key`.\n"
            )

        if default_headers:
            self.default_headers = default_headers

        if http_client:
            self._http_client = http_client
        else:
            self._http_client = Client(timeout=timeout)

        self.chat: resources.Chat = resources.Chat(self)
        self.images: resources.Images = resources.Images(self)
        self.audio: resources.Audio = resources.Audio(self)
        self.moderations: resources.Moderations = resources.Moderations(self)
        self.embeddings: resources.Embeddings = resources.Embeddings(self)
        self.insults: resources.Insults = resources.Insults(self)
        self.jokes: resources.Jokes = resources.Jokes(self)

    def __del__(self) -> None:
        self._http_client.close()

    def _check_response_status_codes(self, response: Response) -> None:
        if response.status_code in {429, 500, 502, 503, 504}:
            raise ShuttleAIAPIStatusException.from_response(
                response,
                message=f"Status: {response.status_code}. Message: {response.text}",
            )
        elif 400 <= response.status_code < 500:
            if response.stream:
                response.read()
            raise ShuttleAIAPIException.from_response(
                response,
                message=f"Status: {response.status_code}. Message: {response.text}",
            )
        elif response.status_code >= 500:
            if response.stream:
                response.read()
            raise ShuttleAIException(
                message=f"Status: {response.status_code}. Message: {response.text}",
            )

    def _check_streaming_response(self, response: Response) -> None:
        self._check_response_status_codes(response)

    def _check_response(self, response: Response) -> Dict[str, Any]:
        self._check_response_status_codes(response)

        json_response: Dict[str, Any] = orjson.loads(response.content)

        return json_response

    def _request(
        self,
        method: str,
        json: Optional[Dict[str, Any]],
        path: str,
        stream: bool = False,
    ) -> Iterator[Dict[str, Any]]:
        json_bytes: Any | None = None
        if json and len(json) > 0:
            if "file" in json:
                with open(json["file"], "rb") as f:
                    json_bytes = {"file": f.read()}
            else:
                json_bytes = orjson.dumps(json)
        # json_bytes: bytes | None = (
        #     orjson.dumps(json) if json and len(json) > 0 else None
        # )  # x-sai [dict to bytes]

        accept_header = "text/event-stream" if stream else "application/json"
        headers = {
            "Accept": accept_header,
            "User-Agent": f"shuttleai-python/{self._version}",
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }

        if self.default_headers:
            headers.update(self.default_headers)

        url = posixpath.join(self.base_url, path)

        self._logger.debug(f"Sending request: {method} {url} {json}")

        try:
            if stream:
                with self._http_client.stream(
                    method,
                    url,
                    headers=headers,
                    json=json_bytes,
                ) as response:
                    self._check_streaming_response(response)

                    for line in response.iter_lines():
                        json_streamed_response = self._process_line(line)
                        if json_streamed_response:
                            yield json_streamed_response

            else:
                response = self._http_client.request(
                    method,
                    url,
                    headers=headers,
                    json=json_bytes,
                )

                yield self._check_response(response)

        except ConnectError as e:
            raise ShuttleAIConnectionException(str(e)) from e
        except RequestError as e:
            raise ShuttleAIException(f"Unexpected exception ({e.__class__.__name__}): {e}") from e
        except JSONDecodeError as e:
            raise ShuttleAIAPIException.from_response(
                response,
                message=f"Failed to decode json body: {response.text}",
            ) from e
        except ShuttleAIAPIStatusException as e:
            raise ShuttleAIAPIStatusException.from_response(response, message=str(e)) from e

    def fetch_model(self, model_id: str) -> BaseModelCard:
        """Fetches a model by its ID

        Args:
            model_id (str): The ID of the model to fetch

        Returns:
            BaseModelCard, None]: The model if it exists
        """
        singleton_response = self._request("get", {}, f"v1/models/{model_id}")
        try:
            return BaseModelCard(**next(singleton_response)["data"])
        except (pydantic_core.ValidationError, StopIteration) as e:
            raise ShuttleAIException("No response received") from e

    def list_models(self) -> Union[ListModelsResponse, ListVerboseModelsResponse]:
        """Returns a list of the available models

        Returns:
            ListModelsResponse: A response object containing the list of models.
        """
        return self._fetch_and_process_models("v1/models", ListModelsResponse)

    def list_models_verbose(
        self,
    ) -> Union[ListVerboseModelsResponse, ListModelsResponse]:
        """Returns a list of the available models with verbose information

        Returns:
            ListVerboseModelsResponse: A response object containing the list of models.
        """
        return self._fetch_and_process_models("v1/models/verbose", ListVerboseModelsResponse)

    def _fetch_and_process_models(
        self,
        endpoint: str,
        response_class: Type[Union[ListModelsResponse, ListVerboseModelsResponse]],
    ) -> Union[ListModelsResponse, ListVerboseModelsResponse]:
        singleton_response = self._request("get", {}, endpoint)
        try:
            list_models_response = response_class(**next(singleton_response))
        except pydantic_core.ValidationError as e:
            raise ShuttleAIException("No response received") from e

        models_by_id = {model.id: model for model in list_models_response.data}

        for model in list_models_response.data:
            if isinstance(model, ProxyCard):
                model_parent = models_by_id.get(model.proxy_to)
                assert isinstance(model_parent, BaseModelCard)
                if model_parent:
                    model.parent = model_parent

        return list_models_response

    @overload
    def ez_chat(  # type: ignore
        self, text: str, model: Optional[str] = None, stream: Literal[False] = False
    ) -> ChatCompletionResponse: ...

    @overload
    def ez_chat(
        self, text: str, model: Optional[str] = None, stream: Literal[True] = True
    ) -> Iterable[ChatCompletionStreamResponse]: ...

    def ez_chat(
        self,
        text: str,
        model: Optional[str] = None,
        stream: bool = False,
    ) -> Union[ChatCompletionResponse, Iterable[ChatCompletionStreamResponse]]:
        return self.chat.completions.create(  # type: ignore
            messages=[{"role": "user", "content": text}],
            model=model,
            stream=stream,
        )
